from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, Literal, Optional, Union

import numpy as np
import pandas as pd

from pyfixest.errors import (
    NonConvergenceError,
)
from pyfixest.estimation.demean_ import demean
from pyfixest.estimation.feols_ import Feols, PredictionErrorOptions, PredictionType
from pyfixest.estimation.fepois_ import _check_for_separation
from pyfixest.estimation.FormulaParser import FixestFormula
from pyfixest.utils.dev_utils import DataFrameType


class Feglm(Feols, ABC):
    "Abstract base class for the estimation of a fixed-effects GLM model."

    def __init__(
        self,
        FixestFormula: FixestFormula,
        data: pd.DataFrame,
        ssc_dict: dict[str, Union[str, bool]],
        drop_singletons: bool,
        drop_intercept: bool,
        weights: Optional[str],
        weights_type: Optional[str],
        collin_tol: float,
        fixef_tol: float,
        fixef_maxiter: int,
        lookup_demeaned_data: dict[str, pd.DataFrame],
        tol: float,
        maxiter: int,
        solver: Literal[
            "np.linalg.lstsq",
            "np.linalg.solve",
            "scipy.linalg.solve",
            "scipy.sparse.linalg.lsqr",
            "jax",
        ],
        store_data: bool = True,
        copy_data: bool = True,
        lean: bool = False,
        sample_split_var: Optional[str] = None,
        sample_split_value: Optional[Union[str, int]] = None,
        separation_check: Optional[list[str]] = None,
        context: Union[int, Mapping[str, Any]] = 0,
    ):
        super().__init__(
            FixestFormula=FixestFormula,
            data=data,
            ssc_dict=ssc_dict,
            drop_singletons=drop_singletons,
            drop_intercept=drop_intercept,
            weights=weights,
            weights_type=weights_type,
            collin_tol=collin_tol,
            fixef_tol=fixef_tol,
            fixef_maxiter=fixef_maxiter,
            lookup_demeaned_data=lookup_demeaned_data,
            solver=solver,
            store_data=store_data,
            copy_data=copy_data,
            lean=lean,
            sample_split_var=sample_split_var,
            sample_split_value=sample_split_value,
            context=context,
        )

        _glm_input_checks(drop_singletons=drop_singletons, tol=tol, maxiter=maxiter)

        self.maxiter = maxiter
        self.tol = tol
        self.convergence = False
        self.separation_check = separation_check

        self._support_crv3_inference = True
        self._support_iid_inference = True
        self._supports_cluster_causal_variance = False
        self._support_decomposition = False

        self._Y_hat_response = np.empty(0)
        self.deviance = None
        self._Xbeta = np.empty(0)

        self._method = "feglm"

    def prepare_model_matrix(self):
        "Prepare model inputs for estimation."
        super().prepare_model_matrix()

        if self._fe is not None:
            raise NotImplementedError("Fixed effects are not yet supported for GLMs.")

        # check for separation
        na_separation: list[int] = []
        if (
            self._fe is not None
            and self.separation_check is not None
            and self.separation_check  # not an empty list
        ):
            na_separation = _check_for_separation(
                Y=self._Y,
                X=self._X,
                fe=self._fe,
                fml=self._fml,
                data=self._data,
                methods=self.separation_check,
            )

        if na_separation:
            self._Y.drop(na_separation, axis=0, inplace=True)
            self._X.drop(na_separation, axis=0, inplace=True)
            self._fe.drop(na_separation, axis=0, inplace=True)
            self._data.drop(na_separation, axis=0, inplace=True)
            self._N = self._Y.shape[0]

            self.na_index = np.concatenate([self.na_index, np.array(na_separation)])
            self.n_separation_na = len(na_separation)

    def to_array(self):
        "Turn estimation DataFrames to np arrays."
        self._Y, self._X, self._Z = (
            self._Y.to_numpy(),
            self._X.to_numpy(),
            self._X.to_numpy(),
        )
        if self._fe is not None:
            self._fe = self._fe.to_numpy()
            if self._fe.ndim == 1:
                self._fe = self._fe.reshape((self._N, 1))

    def get_fit(self):
        """
        Fit the GLM model via iterated weighted least squares.

        All equations in the code refer to Stammann 'Fast and Feasible
        Estimation of Generalized Linear Models with High-Dimensional
        k-way fixed effects': https://arxiv.org/pdf/1707.01815
        """
        _Y = self._Y
        _X = self._X
        _fe = self._fe
        _N = self._N
        _convergence = self.convergence  # False
        _maxiter = self.maxiter
        _tol = self.tol
        _fixef_tol = self._fixef_tol
        _fixef_maxiter = self._fixef_maxiter
        _solver = self._solver

        # initialize

        beta = np.zeros(_X.shape[1])
        eta = np.zeros(_N)
        mu = self._get_mu(theta=eta)
        deviance = self._get_deviance(_Y.flatten(), mu)
        deviance_old = deviance.copy() + 1

        for r in range(_maxiter):
            if r == 0:
                pass
            else:
                converged = self._check_convergence(
                    crit=self._get_diff(deviance=deviance, last=deviance_old),
                    tol=_tol,
                    r=r,
                    maxiter=_maxiter,
                    model=self._method,
                )
                if converged:
                    break

            # Step 1: _get weights w_tilde(r-1) and v(r-1) (eq. 2.5)
            detadmu = self._update_detadmu(mu=mu)
            W = self._update_W(mu=mu)

            # Step 2: _get v_tilde(r-1) and X_tilde(r-1) (eq. 3.2)
            W_tilde = self._update_W_tilde(W=W)
            X_tilde = self._update_X_tilde(W_tilde=W_tilde, X=_X)
            v_tilde = self._update_v_tilde(
                y=_Y.flatten(), mu=mu, W_tilde=W_tilde.flatten(), detadmu=detadmu
            )

            # Step 3 compute v_dotdot(r-1) and X_dotdot(r-1) - demeaning
            v_dotdot, X_dotdot = self.residualize(
                v=v_tilde,
                X=X_tilde,
                flist=_fe,
                weights=W_tilde.flatten(),
                tol=_fixef_tol,
                maxiter=_fixef_maxiter,
            )

            # Step 4: compute (beta(r) - beta(r-1)) and check for convergence, _update beta(r-1) s(eq. 3.5)
            beta_update_diff = self._update_beta_diff(
                X_dotdot=X_dotdot, v_dotdot=v_dotdot
            )

            # Step 5: _update using step halfing (if required)

            deviance_old = deviance.copy()

            beta, eta, mu, deviance = self._update_eta_step_halfing(
                Y=_Y,
                beta=beta,
                eta=eta,
                mu=mu,
                deviance=deviance_old,
                beta_update_diff=beta_update_diff,
                W_tilde=W_tilde.flatten(),
                v_tilde=v_tilde,
                v_dotdot=v_dotdot,
                X_dotdot=X_dotdot,
                deviance_old=deviance_old,
                step_halfing_tolerance=1e-12,
            )

        self._beta_hat = beta.flatten()
        self._Y_hat_response = mu.flatten()
        self._Y_hat_link = eta.flatten()

        # _update for inference
        self._weights = W
        self._irls_weights = W

        # if only one dim
        if self._weights.ndim == 1:
            self._weights = self._weights.reshape((self._N, 1))

        self._u_hat_response = (self._Y.flatten() - self._get_mu(theta=eta)).flatten()
        self._u_hat_working = (
            self._u_hat_response
            if self._method == "feglm-gaussian"
            else (v_dotdot / W_tilde).flatten()
        )

        self._scores_response = self._u_hat_response[:, None] * self._X
        self._scores_working = self._u_hat_working[:, None] * self._X

        self._scores = self._get_score(y=_Y.flatten(), X=_X, mu=mu, eta=eta)

        self._u_hat = self._u_hat_working
        self._tZX = np.transpose(self._Z) @ self._X
        self._tZXinv = np.linalg.inv(self._tZX)
        self._Xbeta = eta

        self._hessian = X_dotdot.T @ X_dotdot
        self.deviance = deviance

        if _convergence:
            self._convergence = True

    def _vcov_iid(self):
        return self._bread

    def _update_v(
        self, y: np.ndarray, mu: np.ndarray, detadmu: np.ndarray
    ) -> np.ndarray:
        "Get (running) dependent variable v for the GLM family."
        return (y - mu) * detadmu

    def _update_W(self, mu: np.ndarray) -> np.ndarray:
        "Get (running) weights W for the GLM family."
        return 1 / (self._update_detadmu(mu=mu) ** 2 * self._get_V(mu=mu))

    def _update_W_tilde(self, W: np.ndarray) -> np.ndarray:
        "Get W_tilde (formula 3.2)."
        return np.sqrt(W)

    def _update_v_tilde(
        self, y: np.ndarray, mu: np.ndarray, W_tilde: np.ndarray, detadmu: np.ndarray
    ) -> np.ndarray:
        "Get v_tilde (formula 3.2)."
        return W_tilde * ((y - mu) * detadmu)

    def _update_X_tilde(self, W_tilde: np.ndarray, X: np.ndarray) -> np.ndarray:
        "Get X_tilde (formula 3.2)."
        return W_tilde.reshape(-1, 1) * X

    def _update_beta_diff(
        self, X_dotdot: np.ndarray, v_dotdot: np.ndarray
    ) -> np.ndarray:
        "Get the beta _update difference (formula 3.5) via WLS fit."
        beta_diff = np.linalg.lstsq(X_dotdot, v_dotdot.reshape(-1, 1), rcond=None)[
            0
        ].flatten()
        return beta_diff

    def _update_eta(
        self,
        W_tilde: np.ndarray,
        v_tilde: np.ndarray,
        v_dotdot: np.ndarray,
        X_dotdot: np.ndarray,
        beta_diff: np.ndarray,
        eta: np.ndarray,
    ) -> np.ndarray:
        "Get the eta _update (formula 4.5)."
        return eta + X_dotdot @ beta_diff / W_tilde

    def _get_gradient(self, Z: np.ndarray, W: np.ndarray, v: np.ndarray) -> np.ndarray:
        return Z.T @ W @ v

    def _get_diff(self, deviance: np.ndarray, last: np.ndarray) -> np.ndarray:
        return np.abs(deviance - last) / (0.1 + np.abs(last))

    def residualize(
        self,
        v: np.ndarray,
        X: np.ndarray,
        flist: np.ndarray,
        weights: np.ndarray,
        tol: np.ndarray,
        maxiter: int,
    ) -> tuple[np.ndarray, np.ndarray]:
        "Residualize v and X by flist using weights."
        if flist is None:
            return v, X
        else:
            vX_resid, success = demean(
                x=np.c_[v, X], flist=flist, weights=weights, tol=tol, maxiter=maxiter
            )
            if success is False:
                raise ValueError(f"Demeaning failed after {maxiter} iterations.")
            else:
                return vX_resid[:, 0], vX_resid[:, 1:]

    def _check_convergence(
        self,
        crit: float,
        tol: float,
        r: int,
        maxiter: int,
        model: str,
    ) -> bool:
        if model == "feglm-gaussian":
            converged = True
        else:
            converged = crit < tol
            if r == maxiter:
                raise NonConvergenceError(
                    f"""
                    The IRLS algorithm did not converge with {maxiter}
                    iterations. Try to increase the maximum number of iterations.
                    """
                )

        return converged

    def _update_eta_step_halfing(
        self,
        Y: np.ndarray,
        beta: np.ndarray,
        eta: np.ndarray,
        mu: np.ndarray,
        deviance: np.ndarray,
        beta_update_diff: np.ndarray,
        W_tilde: np.ndarray,
        v_tilde: np.ndarray,
        v_dotdot: np.ndarray,
        X_dotdot: np.ndarray,
        deviance_old: np.ndarray,
        step_halfing_tolerance: float,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        "Update parameters, potentially using step halfing."
        alpha = 1.0
        step_accepted = False

        while alpha > step_halfing_tolerance:
            beta_try = beta + alpha * beta_update_diff
            eta_try = self._update_eta(
                W_tilde=W_tilde.flatten(),
                v_tilde=v_tilde,
                v_dotdot=v_dotdot,
                X_dotdot=X_dotdot,
                beta_diff=alpha * beta_update_diff,
                eta=eta,
            )
            mu_try = self._get_mu(theta=eta_try)
            deviance_try = self._get_deviance(Y.flatten(), mu_try)
            if deviance_try < deviance_old:
                beta = beta_try
                eta = eta_try
                mu = mu_try
                deviance = deviance_try
                step_accepted = True
                break
            else:
                alpha /= 2.0

        if not step_accepted:
            raise RuntimeError("Step-halving failed to find improvement.")

        return beta, eta, mu, deviance

    def predict(
        self,
        newdata: Optional[DataFrameType] = None,
        atol: float = 1e-6,
        btol: float = 1e-6,
        type: PredictionType = "link",
        se_fit: Optional[bool] = False,
        interval: Optional[PredictionErrorOptions] = None,
        alpha: float = 0.05,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """
        Return predicted values from regression model.

        Return a flat np.array with predicted values of the regression model.
        If new fixed effect levels are introduced in `newdata`, predicted values
        for such observations
        will be set to NaN.

        Parameters
        ----------
        newdata : Union[None, pd.DataFrame], optional
            A pd.DataFrame with the new data, to be used for prediction.
            If None (default), uses the data used for fitting the model.
        atol : Float, default 1e-6
            Stopping tolerance for scipy.sparse.linalg.lsqr().
            See https://docs.scipy.org/doc/
                scipy/reference/generated/scipy.sparse.linalg.lsqr.html
        btol : Float, default 1e-6
            Another stopping tolerance for scipy.sparse.linalg.lsqr().
            See https://docs.scipy.org/doc/
                scipy/reference/generated/scipy.sparse.linalg.lsqr.html
        type : str, optional
            The type of prediction to be computed.
            Can be either "response" (default) or "link".
            If type="response", the output is at the level of the response variable,
            i.e., it is the expected predictor E(Y|X).
            If "link", the output is at the level of the explanatory variables,
            i.e., the linear predictor X @ beta.
        atol : Float, default 1e-6
            Stopping tolerance for scipy.sparse.linalg.lsqr().
            See https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsqr.html
        btol : Float, default 1e-6
            Another stopping tolerance for scipy.sparse.linalg.lsqr().
            See https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsqr.html
        se_fit: Optional[bool], optional
            If True, the standard error of the prediction is computed. Only feasible
            for models without fixed effects. GLMs are not supported. Defaults to False.
        interval: str, optional
            The type of interval to compute. Can be either 'prediction' or None.
        alpha: float, optional
            The alpha level for the confidence interval. Defaults to 0.05. Only
            used if interval = "prediction" is not None.

        Returns
        -------
        Union[np.ndarray, pd.DataFrame]
            Returns a pd.Dataframe with columns "fit", "se_fit" and CIs if argument "interval=prediction".
            Otherwise, returns a np.ndarray with the predicted values of the model or the prediction
            standard errors if argument "se_fit=True".
        """
        if se_fit:
            raise NotImplementedError(
                "Prediction with standard errors is not implemented for GLMs."
            )

        yhat = super().predict(newdata=newdata, type="link", atol=atol, btol=btol)
        if type == "response":
            return self._get_mu(
                theta=yhat.to_numpy() if isinstance(yhat, pd.DataFrame) else yhat
            )
        else:
            return yhat

    @abstractmethod
    def _check_dependent_variable(self):
        pass

    @abstractmethod
    def _get_score(
        self, y: np.ndarray, X: np.ndarray, mu: np.ndarray, eta: np.ndarray
    ) -> np.ndarray:
        pass

    @abstractmethod
    def _get_deviance(self, y: np.ndarray, mu: np.ndarray) -> np.ndarray:
        "Compute the deviance for the GLM family."
        pass

    @abstractmethod
    def _get_dispersion_phi(self, theta: np.ndarray) -> float:
        "Get the dispersion parameter phi for the GLM family."
        pass

    @abstractmethod
    def _get_b(self, theta: np.ndarray) -> np.ndarray:
        "Get the cumulant function b(theta) for the GLM family."
        pass

    @abstractmethod
    def _get_mu(self, theta: np.ndarray) -> np.ndarray:
        "Get the mean mu(theta) for the GLM family."
        pass

    @abstractmethod
    def _get_link(self, mu: np.ndarray) -> np.ndarray:
        "Get the link function theta(mu) for the GLM family."
        pass

    @abstractmethod
    def _update_detadmu(self, mu: np.ndarray) -> np.ndarray:
        "Get the derivative of mu(theta) with respect to theta for the GLM family."
        pass

    @abstractmethod
    def _get_theta(self, mu: np.ndarray) -> np.ndarray:
        "Get the mechanical link theta(mu) for the GLM family."
        pass

    @abstractmethod
    def _get_V(self, mu: np.ndarray) -> np.ndarray:
        "Get the variance function V(mu) for the GLM family."
        pass


def _glm_input_checks(drop_singletons: bool, tol: float, maxiter: int):
    """
    Perform input checks for Fepois constructor arguments.

    Parameters
    ----------
    drop_singletons : bool
        Whether to drop singleton fixed effects.
    tol : float
        Tolerance level for convergence check.
    maxiter : int
        Maximum number of iterations.

    Returns
    -------
    None
    """
    # drop singletons must be logical
    if not isinstance(drop_singletons, bool):
        raise TypeError("drop_singletons must be logical.")
    # tol must be numeric and between 0 and 1
    if not isinstance(tol, (int, float)):
        raise TypeError("tol must be numeric.")
    if tol <= 0 or tol >= 1:
        raise AssertionError("tol must be between 0 and 1.")
    # maxiter must be integer and greater than 0
    if not isinstance(maxiter, int):
        raise TypeError("maxiter must be integer.")
    if maxiter <= 0:
        raise AssertionError("maxiter must be greater than 0.")



import warnings
from collections.abc import Mapping
from importlib import import_module
from typing import Any, Literal, Optional, Union

import numpy as np
import pandas as pd

from pyfixest.estimation.demean_ import demean_model
from pyfixest.estimation.feols_ import Feols, _drop_multicollinear_variables
from pyfixest.estimation.FormulaParser import FixestFormula
from pyfixest.estimation.literals import DemeanerBackendOptions
from pyfixest.estimation.solvers import solve_ols


class Feiv(Feols):
    """
    Non user-facing class to estimate an IV model using a 2SLS estimator.

    Inherits from the Feols class. Users should not directly instantiate this class,
    but rather use the [feols()](/reference/estimation.feols.qmd) function. Note that
    no demeaning is performed in this class: demeaning is performed in the
    [FixestMulti](/reference/estimation.fixest_multi.qmd) class (to allow for caching
    of demeaned variables for multiple estimation).

    Parameters
    ----------
    Y : np.ndarray
        Dependent variable, a two-dimensional np.array.
    X : np.ndarray
        Independent variables, a two-dimensional np.array.
    endgvar : np.ndarray
        Endogenous Indenpendent variables, a two-dimensional np.array.
    Z : np.ndarray
        Instruments, a two-dimensional np.array.
    weights : np.ndarray
        Weights, a one-dimensional np.array.
    coefnames_x : list
        Names of the coefficients of X.
    coefnames_z : list
        Names of the coefficients of Z.
    collin_tol : float
        Tolerance for collinearity check.
    solver: Literal["np.linalg.lstsq", "np.linalg.solve", "scipy.linalg.solve",
        "scipy.sparse.linalg.lsqr", "jax"],
        default is "scipy.linalg.solve". Solver to use for the estimation.
    demeaner_backend: DemeanerBackendOptions, optional
        The backend to use for demeaning. Can be either "numba", "jax", or "rust".
        Defaults to "numba".
    weights_name : Optional[str]
        Name of the weights variable.
    weights_type : Optional[str]
        Type of the weights variable. Either "aweights" for analytic weights
        or "fweights" for frequency weights.

    Attributes
    ----------
    _Z : np.ndarray
        Processed instruments after handling multicollinearity.
    _weights_type_feiv : str
        Type of the weights variable defined in Feiv class.
        Either "aweights" for analytic weights or "fweights"
        for frequency weights.
    _coefnames_z : list
        Names of coefficients for Z after handling multicollinearity.
    _collin_vars_z : list
        Variables identified as collinear in Z.
    _collin_index_z : list
        Indices of collinear variables in Z.
    _is_iv : bool
        Indicator if instrumental variables are used.
    _support_crv3_inference : bool
        Indicator for supporting CRV3 inference.
    _support_iid_inference : bool
        Indicator for supporting IID inference.
    _tZX : np.ndarray
        Transpose of Z times X.
    _tXZ : np.ndarray
        Transpose of X times Z.
    _tZy : np.ndarray
        Transpose of Z times Y.
    _tZZinv : np.ndarray
        Inverse of transpose of Z times Z.
    _beta_hat : np.ndarray
        Estimated regression coefficients.
    _Y_hat_link : np.ndarray
        Predicted values of the regression model.
    _u_hat : np.ndarray
        Residuals of the regression model.
    _scores : np.ndarray
        Scores used in the regression.
    _hessian : np.ndarray
        Hessian matrix used in the regression.
    _bread : np.ndarray
        Bread matrix used in the regression.
    _pi_hat : np.ndarray
        Estimated coefficients from 1st stage regression
    _X_hat : np.ndarray
        Predicted values of the 1st stage regression
    _v_hat : np.ndarray
        Residuals of the 1st stage regression
    _model_1st_stage : Any
        feols object of 1st stage regression.
        It contains various results and diagnostics
        from the fixed effects OLS regression.
    _endogvar_1st_stage : np.ndarray
        Unweihgted Endogenous independent variable vector
    _Z_1st_stage : np.ndarray
        Unweighted instruments vector to be used for 1st stage
    _non_exo_instruments : list
        List of instruments name excluding exogenous independent vars.
    __p_iv : scalar
        Number of instruments listed in _non_exo_instruments
    _f_stat_1st_stage : scalar
        F-statistics of First Stage regression for evaluation of IV weakness.
        The computed F-statistics test the following null hypothesis :
        # H0 : β_{z_1} = 0 & ... & β_{z_{p_iv}} = 0 where z_1, ..., z_{p_iv}
        # are the instrument variables
        # H1 : H0 does not hold
        Note that this F-statistics is adjusted to heteroskedasticity /
        clusters if users set specification of variance-covariance matrix type
    _eff_F : scalar
        Effective F-statistics of first stage regression as in Olea and Pflueger 2013
    _data: pd.DataFrame
        The data frame used in the estimation. None if arguments `lean = True` or
        `store_data = False`.


    Raises
    ------
    ValueError
        If Z is not a two-dimensional array.

    """

    # Constructor and methods implementation...
    def __init__(
        self,
        FixestFormula: FixestFormula,
        data: pd.DataFrame,
        ssc_dict: dict[str, Union[str, bool]],
        drop_singletons: bool,
        drop_intercept: bool,
        weights: Optional[str],
        weights_type: Optional[str],
        collin_tol: float,
        fixef_tol: float,
        fixef_maxiter: int,
        lookup_demeaned_data: dict[str, pd.DataFrame],
        solver: Literal[
            "np.linalg.lstsq",
            "np.linalg.solve",
            "scipy.linalg.solve",
            "scipy.sparse.linalg.lsqr",
            "jax",
        ] = "scipy.linalg.solve",
        demeaner_backend: DemeanerBackendOptions = "numba",
        store_data: bool = True,
        copy_data: bool = True,
        lean: bool = False,
        context: Union[int, Mapping[str, Any]] = 0,
        sample_split_var: Optional[str] = None,
        sample_split_value: Optional[Union[str, int]] = None,
    ) -> None:
        super().__init__(
            FixestFormula=FixestFormula,
            data=data,
            ssc_dict=ssc_dict,
            drop_singletons=drop_singletons,
            drop_intercept=drop_intercept,
            weights=weights,
            weights_type=weights_type,
            collin_tol=collin_tol,
            fixef_tol=fixef_tol,
            fixef_maxiter=fixef_maxiter,
            lookup_demeaned_data=lookup_demeaned_data,
            solver=solver,
            store_data=store_data,
            copy_data=copy_data,
            lean=lean,
            sample_split_var=sample_split_var,
            sample_split_value=sample_split_value,
            context=context,
            demeaner_backend=demeaner_backend,
        )

        self._is_iv = True
        self._support_crv3_inference = False
        self._support_iid_inference = True
        self._supports_cluster_causal_variance = False
        self._support_decomposition = False

    def wls_transform(self) -> None:
        "Transform variables for WLS estimation."
        super().wls_transform()
        if self._has_weights:
            w = np.sqrt(self._weights)
            self._endogvar = self._endogvar * w
            self._Z = self._Z * w

    def to_array(self) -> None:
        "Transform estimation DataFrames to arrays."
        super().to_array()
        self._Z = self._Zd.to_numpy()
        self._endogvar = self._endogvar.to_numpy()

    def demean(self) -> None:
        "Demean instruments and endogeneous variable."
        super().demean()
        if self._has_fixef:
            self._endogvard, self._Zd = demean_model(
                self._endogvar,
                self._Z,
                self._fe,
                self._weights.flatten(),
                self._lookup_demeaned_data,
                self._na_index_str,
                self._fixef_tol,
                self._fixef_maxiter,
                self._demean_func,
            )
        else:
            self._endogvard = self._endogvar
            self._Zd = self._Z

    def drop_multicol_vars(self) -> None:
        "Drop multicollinear variables in matrix of instruments Z."
        super().drop_multicol_vars()
        (
            self._Z,
            self._coefnames_z,
            self._collin_vars_z,
            self._collin_index_z,
        ) = _drop_multicollinear_variables(
            self._Z,
            self._coefnames_z,
            self._collin_tol,
            self._find_collinear_variables_func,
        )

    def get_fit(self) -> None:
        """Fit a IV model using a 2SLS estimator."""
        _X = self._X
        _Z = self._Z
        _Y = self._Y

        _solver = self._solver

        # Start Second Stage
        self._tZX = _Z.T @ _X
        self._tXZ = _X.T @ _Z
        self._tZy = _Z.T @ _Y
        self._tZZinv = np.linalg.inv(_Z.T @ _Z)

        H = self._tXZ @ self._tZZinv
        A = H @ self._tZX
        B = H @ self._tZy
        self._beta_hat = solve_ols(A, B, _solver)

        # residuals
        self._u_hat = self._Y.flatten() - (self._X @ self._beta_hat).flatten()
        self._get_predictors()

        # Compute scores and hessian
        self._scores = self._Z * self._u_hat[:, None]
        self._hessian = self._Z.T @ self._Z

        # Compute bread matrix
        D = np.linalg.inv(self._tXZ @ self._tZZinv @ self._tZX)
        self._bread = H.T @ D @ H

    def first_stage(self) -> None:
        """Implement First stage regression."""
        # Store names of instruments from Z matrix
        self._non_exo_instruments = list(set(self._coefnames_z) - set(self._coefnames))

        fixest_module = import_module("pyfixest.estimation")
        fit_ = fixest_module.feols

        fml_first_stage = self.FixestFormula.fml_first_stage.replace(" ", "")
        if self._has_fixef:
            fml_first_stage += f" | {self._fixef}"

        # Type hint to reflect that vcov_detail can be either a dict or a str
        vcov_detail: Union[dict[str, str], str]

        if self._is_clustered:
            a = self._clustervar[0]
            vcov_detail = {self._vcov_type_detail: a}
        else:
            vcov_detail = self._vcov_type_detail

        # Do first stage regression
        model1 = fit_(
            fml=fml_first_stage,
            data=self._data,
            vcov=vcov_detail,
            weights=self._weights_name,
            weights_type=self._weights_type,
            collin_tol=self._collin_tol,
        )

        # Ensure model1 is of type Feols
        if isinstance(model1, Feols):
            # Store the first stage coefficients
            self._pi_hat = model1._beta_hat

            # Use fitted values from the first stage
            self._X_hat = (
                model1._X @ model1._beta_hat
            )  # note that model1._X is demeaned

            # Residuals from the first stage
            self._v_hat = model1._u_hat

            # Store 1st stage model for further use
            self._model_1st_stage = model1

        else:
            raise TypeError("The first stage model must be of type Feols")

        self.IV_weakness_test(["f_stat"])

    def IV_Diag(self, statistics: Optional[list[str]] = None):
        """Implement IV diagnostic tests.

        Notes
        -----
        This method covers diagnostic tests related with IV regression.
        We currently have IV weak tests only. More test will be updated
        in future updates!

        Parameters
        ----------
        statistics : list[str], optional
            List of IV diagnostic statistics

        Example
        -------
        The following is an example usage of this method:

            ```{python}

            import numpy as np
            import pandas as pd
            from pyfixest.estimation.estimation import feols

            # Set random seed for reproducibility
            np.random.seed(1)

            # Number of observations
            n = 1000

            # Simulate the data
            # Instrumental variable
            z = np.random.binomial(1, 0.5, size=n)
            z2 = np.random.binomial(1, 0.5, size=n)

            # Endogenous variable
            d = 0.5 * z + 1.5 * z2 + np.random.normal(size=n)

            # Control variables
            c1 = np.random.normal(size=n)
            c2 = np.random.normal(size=n)

            # Outcome variable
            y = 1.0 + 1.5 * d + 0.8 * c1 + 0.5 * c2 + np.random.normal(size=n)

            # Cluster variable
            cluster = np.random.randint(1, 50, size=n)
            weights = np.random.uniform(1, 3, size=n)

            # Create a DataFrame
            data = pd.DataFrame({
                'd': d,
                'y': y,
                'z': z,
                'z2': z2,
                'c1': c1,
                'c2': c2,
                'cluster': cluster,
                'weights': weights
            })

            vcov_detail = "iid"

            # Fit OLS model
            fit_ols = feols("y ~ 1 + d + c1 + c2", data=data, vcov=vcov_detail)

            # Fit IV model
            fit_iv = feols("y ~ 1 + c1 + c2 | d ~ z", data=data,
                     vcov=vcov_detail,
                     weights="weights")
            fit_iv.first_stage()
            F_stat_pf = fit_iv._f_stat_1st_stage
            fit_iv.IV_Diag()
            F_stat_eff_pf = fit_iv._eff_F

            print("(Unadjusted) F stat :", F_stat_pf)
            print("Effective F stat :", F_stat_eff_pf)

            ```
        """
        # Set default statistics
        iv_diag_stat = ["f_stat", "effective_f"]

        # Set statistics allowed in the current version
        iv_diag_stat_allowed = ["f_stat", "effective_f"]

        # Check whether there is unsupported statistics.
        if statistics:
            invalid_stats = [
                stat for stat in statistics if stat not in iv_diag_stat_allowed
            ]

            if invalid_stats:
                raise ValueError(
                    f"Statistics not supported: {invalid_stats}."
                    f"You should specify from the following list of statistics {iv_diag_stat_allowed}"
                )

            iv_diag_stat += statistics

        self.IV_weakness_test(iv_diag_stat)

    def IV_weakness_test(self, iv_diag_statistics: Optional[list[str]] = None) -> None:
        """Implement IV weakness test (F-test).

        This method covers hetero-robust and clustered-robust F statistics.
        It produces two statistics:

        - self._f_stat_1st_stage: F statistics of first stage regression
        - self._eff_F: Effective F statistics (Olea and Pflueger 2013)
                       of first stage regression

        Notes
        -----
        "self._f_stat_1st_stage" is adjusted to the specification of vcov.
        If vcov_detail = "iid", F statistics is not adjusted,
        otherwise it is always adjusted.

        Parameters
        ----------
        iv_diag_statistics : list, optional
            List of IV weakness statistics

        """
        iv_diag_statistics = iv_diag_statistics or []

        if "f_stat" in iv_diag_statistics:
            self._p_iv = len(self._non_exo_instruments)

            # Create an identity matrix of size p_iv by p_iv
            # Pad the identity matrix with zeros to make it of size p_iv by k
            p_iv = self._p_iv  # number of IVs
            k = (
                self._model_1st_stage._k
            )  # number of estimated coefficients of 1st stage

            # Extract all the IV indexes and its first index
            self._iv_loc = [
                self._coefnames_z.index(x)
                for x in self._non_exo_instruments
                if x in self._coefnames_z
            ]

            # Generate matrix R that tests the following;
            # H0 : \beta_{z_1} = 0 & ... & \beta_{z_{p_iv}} = 0
            #      where z_1, ..., z_{p_iv} are the instrument variables
            # H1 : H0 does not hold

            # Pad identity matrix to implement wald-test
            R = np.zeros((p_iv, k))
            R[:, self._iv_loc] = np.eye(p_iv)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._model_1st_stage.wald_test(R=R)
            self._f_stat_1st_stage = self._model_1st_stage._f_statistic
            self._p_value_1st_stage = self._model_1st_stage._p_value

        if "effective_f" in iv_diag_statistics:
            self.eff_F()

    def eff_F(self) -> None:
        """Compute Effective F stat (Olea and Pflueger 2013)."""
        # If vcov is iid, redo first stage regression

        if self._vcov_type_detail == "iid":
            self._vcov_type_detail = "hetero"
            self._model_1st_stage.vcov("hetero")

        # Compute Effective F stat by Olea and Pflueger 2013
        # 1. Extract First Stage Coefficients and Variance-Covariance Matrix:
        #   Extract the coefficients for the instrument z
        #   from the first stage regression.
        #   Extract the robust variance-covariance matrix of these coefficients.
        #   Extract the instrument matrix.
        # 2. Compute the Instrument Matrix:
        #   Construct the instrument matrix Q_{zz} = Z.T x Z
        # 3. Compute the Effective F-statistic:
        #   F_{eff} = π.T Q_{zz} π / trance(ΣQ_{zz})

        # Extract coefficients for the non-exogenous instruments

        pi_hat = np.array(self._model_1st_stage.coef()[self._non_exo_instruments])
        iv_positions = [
            self._coefnames_z.index(instrument)
            for instrument in self._non_exo_instruments
        ]
        Z = self._model_1st_stage._X[:, iv_positions]

        # Calculate the cross-product of the instrument matrix
        Q_zz = Z.T @ Z

        # Extract the robust variance-covariance matrix
        vcv = self._model_1st_stage._vcov

        # Map the instrument names to their indices in the parameter list
        # Number of rows/columns in vcv

        # Extract the submatrix
        Sigma = vcv[np.ix_(self._iv_loc, self._iv_loc)]

        # Calculate the effective F-statistic
        self._eff_F = (pi_hat.T @ Q_zz @ pi_hat) / np.sum(np.diag(Sigma @ Q_zz))



import numpy as np
import pandas as pd
import pytest
import rpy2.robjects as ro
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import importr

import pyfixest as pf
from pyfixest.estimation.estimation import feols
from pyfixest.estimation.multcomp import _get_rwolf_pval, bonferroni, rwolf
from pyfixest.utils.check_r_install import check_r_install
from pyfixest.utils.utils import get_data

pandas2ri.activate()

# Core R packages
fixest = importr("fixest")
stats = importr("stats")
broom = importr("broom")
# Extended R packages
if import_check := check_r_install("wildrwolf", strict=False):
    wildrwolf = importr("wildrwolf")


@pytest.mark.against_r_core
@pytest.mark.extended
def test_bonferroni():
    data = get_data().dropna()
    rng = np.random.default_rng(989)
    data = get_data()
    data["Y2"] = data["Y"] * rng.normal(0.2, 1, size=len(data))
    data["Y3"] = data["Y2"] + rng.normal(0, 0.5, size=len(data))

    # test set 1

    fit1 = feols("Y ~ X1", data=data)
    fit2 = feols("Y2 ~ X1", data=data)
    fit3 = feols("Y3 ~ X1", data=data)

    bonferroni_py = bonferroni([fit1, fit2, fit3], "X1")

    # R
    fit1_r = fixest.feols(ro.Formula("Y ~ X1"), data=data)
    fit2_r = fixest.feols(ro.Formula("Y2 ~ X1"), data=data)
    fit3_r = fixest.feols(ro.Formula("Y3 ~ X1"), data=data)

    pvalues_r = np.zeros(3)
    for i, x in enumerate([fit1_r, fit2_r, fit3_r]):
        df_tidy = broom.tidy_fixest(x)
        df_r = pd.DataFrame(df_tidy).T
        df_r.columns = ["term", "estimate", "std.error", "statistic", "p.value"]
        pvalues_r[i] = df_r.set_index("term").xs("X1")["p.value"]

    bonferroni_r = stats.p_adjust(pvalues_r, method="bonferroni")

    assert np.all(np.abs(bonferroni_py.iloc[6].values - bonferroni_r) < 0.01), (
        "bonferroni failed"
    )


@pytest.mark.skipif(import_check is False, reason="R package wildrwolf not installed.")
@pytest.mark.against_r_extended
@pytest.mark.extended
@pytest.mark.parametrize("seed", [293, 912, 831])
@pytest.mark.parametrize("sd", [0.5, 1.0, 1.5])
def test_wildrwolf_hc(seed, sd):
    rng = np.random.default_rng(seed)
    data = get_data(N=1_000, seed=seed)
    data["f1"] = rng.choice(range(100), len(data), True)
    data["Y2"] = data["Y"] * rng.normal(0, sd, size=len(data))
    data["Y3"] = data["Y2"] + rng.normal(0, sd, size=len(data))

    # test set 1

    fit = feols("Y + Y2 + Y3~ X1", data=data)
    rwolf_py = rwolf(fit.to_list(), "X1", reps=9999, seed=seed + 2)

    # R
    fit_r = fixest.feols(ro.Formula("c(Y, Y2, Y3) ~ X1"), data=data)
    rwolf_r = wildrwolf.rwolf(fit_r, param="X1", B=9999, seed=seed + 2)

    try:
        np.testing.assert_allclose(
            rwolf_py.iloc[6].values,
            pd.DataFrame(rwolf_r).iloc[5].values.astype(float),
            rtol=0,
            atol=0.01,
            err_msg="rwolf 1 failed",
        )
    except AssertionError:
        rwolf_py = rwolf(fit.to_list(), "X1", reps=29999, seed=seed + 2)
        rwolf_r = wildrwolf.rwolf(fit_r, param="X1", B=29999, seed=seed + 2)

        np.testing.assert_allclose(
            rwolf_py.iloc[6].values,
            pd.DataFrame(rwolf_r).iloc[5].values.astype(float),
            rtol=0,
            atol=0.01,
            err_msg="rwolf 1 failed",
        )


@pytest.mark.skipif(import_check is False, reason="R package wildrwolf not installed.")
@pytest.mark.against_r_extended
@pytest.mark.extended
@pytest.mark.parametrize("seed", [9391])
@pytest.mark.parametrize("sd", [0.5, 1.5])
def test_wildrwolf_crv(seed, sd):
    rng = np.random.default_rng(seed)
    data = get_data(N=4_000, seed=seed)
    data["f1"] = rng.choice(range(100), len(data), True)
    data["Y2"] = data["Y"] * rng.normal(0, sd, size=len(data))
    data["Y3"] = data["Y2"] + rng.normal(0, sd, size=len(data))

    # test set 2

    fit = feols("Y + Y2 + Y3 ~ X1 | f1", data=data)

    rwolf_py = rwolf(fit.to_list(), "X1", reps=9999, seed=seed + 3)

    # R
    fit_r = fixest.feols(ro.Formula("c(Y, Y2, Y3) ~ X1 | f1"), data=data)
    rwolf_r = wildrwolf.rwolf(fit_r, param="X1", B=9999, seed=seed + 3)

    try:
        np.testing.assert_allclose(
            rwolf_py.iloc[6].values,
            pd.DataFrame(rwolf_r).iloc[5].values.astype(float),
            rtol=0,
            atol=0.025,
            err_msg="rwolf 2 failed",
        )

    except AssertionError:
        rwolf_py = rwolf(fit.to_list(), "X1", reps=19999, seed=seed + 3)
        rwolf_r = wildrwolf.rwolf(fit_r, param="X1", B=19999, seed=seed + 3)

        np.testing.assert_allclose(
            rwolf_py.iloc[6].values,
            pd.DataFrame(rwolf_r).iloc[5].values.astype(float),
            rtol=0,
            atol=0.025,
            err_msg="rwolf 2 failed",
        )


@pytest.mark.skipif(import_check is False, reason="R package wildrwolf not installed.")
@pytest.mark.against_r_extended
@pytest.mark.extended
def test_stepwise_function():
    B = 1000
    S = 5

    rng = np.random.default_rng(33)
    t_stat = rng.normal(0, 1, size=S)
    t_boot = rng.normal(0, 1, size=(B, S))

    stepwise_py = _get_rwolf_pval(t_stat, t_boot)
    stepwise_r = wildrwolf.get_rwolf_pval(t_stat, t_boot)

    np.testing.assert_allclose(stepwise_py, stepwise_r)


# Import data from pyfixest


@pytest.mark.extended
@pytest.mark.parametrize("seed", [453])
@pytest.mark.parametrize("reps", [499])
def test_sampling_scheme(seed, reps):
    # Compare RW adjusted p-values from RI and WB resampling methods
    # The p-values should be "largely" similar regardless of resampling methods

    data = get_data().dropna()
    rng = np.random.default_rng(seed)

    # Perturb covariates(not treatment variable)
    data["Y2"] = data["Y"] * rng.normal(0.2, 1, size=len(data))
    data["X2"] = rng.normal(size=len(data))

    fit1 = feols("Y ~ X1 + X2", data=data)
    fit2 = feols("Y ~ X1 + X2 + Y2", data=data)

    # Run rwolf with "ri" sampling method
    rwolf_df_ri = rwolf([fit1, fit2], "X1", reps=reps, seed=seed, sampling_method="ri")

    ri_pval = rwolf_df_ri["est1"]["RW Pr(>|t|)"]

    # Run rwolf with "wild-bootstrap" sampling method
    rwolf_df_wb = rwolf(
        [fit1, fit2], "X1", reps=reps, seed=seed, sampling_method="wild-bootstrap"
    )

    wb_pval = rwolf_df_wb["est1"]["RW Pr(>|t|)"]

    # Calculate percentage difference in p-values
    percent_diff = 100 * (wb_pval - ri_pval) / ri_pval

    print(
        f"Percentage difference in p-values (seed={seed}, reps={reps}): {percent_diff}"
    )

    # Assert that the percentage difference is within an acceptable range
    assert np.abs(percent_diff) < 1.0, (
        f"Percentage difference is too large: {percent_diff}%"
    )


@pytest.mark.extended
def test_multi_vs_list():
    "Test that lists of models and FixestMulti input produce identical results."
    seed = 1232
    reps = 100
    data = pf.get_data(N=100)

    fit_all = pf.feols("Y + Y2 ~ X1 + X2", data=data)
    fit1 = pf.feols("Y ~ X1 + X2", data=data)
    fit2 = pf.feols("Y2 ~ X1 + X2", data=data)

    assert bonferroni(fit_all, "X1").equals(bonferroni([fit1, fit2], "X1"))
    assert rwolf(fit_all, "X1", seed=seed, reps=reps).equals(
        rwolf([fit1, fit2], "X1", seed=seed, reps=reps)
    )


@pytest.mark.extended
@pytest.mark.parametrize("fml", ["Y ~ X1", "Y ~ X1 + X2"])
@pytest.mark.parametrize("seed", [199])
@pytest.mark.parametrize("sampling_method", ["ri", "wild-bootstrap"])
def test_rwolf_vs_wyoung(fml, seed, sampling_method):
    data = pf.get_data(N=100)
    fml1 = fml
    fml2 = f"{fml1} + f1"

    fit1 = pf.feols(fml1, data=data)
    fit2 = pf.feols(fml2, data=data)

    rwolf_output = pf.rwolf(
        [fit1, fit2], "X1", reps=99, seed=seed, sampling_method=sampling_method
    )
    wyoung_output = pf.wyoung(
        [fit1, fit2], "X1", reps=99, seed=seed, sampling_method=sampling_method
    )

    # test that the two pandas dfs are close
    assert np.allclose(rwolf_output, wyoung_output, atol=0.01)



import warnings
from collections.abc import Mapping
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from formulaic import Formula
from scipy.stats import t

from pyfixest.utils.dev_utils import (
    _narwhals_to_pandas,
)


def get_design_matrix_and_yhat(
    model,
    newdata: Optional[pd.DataFrame] = None,
    context: Optional[Union[int, Mapping[str, Any]]] = None,
):
    """
    Build the design matrix X and initializes y_hat for predictions.

    Parameters
    ----------
    model : Feols
        The fitted Feols model (self inside Feols).
    newdata : Optional DataFrame
        The new data on which predictions are made, or None for original data.

    Returns
    -------
    X : np.ndarray or None
        The design matrix. None if X is empty.
    y_hat : np.ndarray
        The (initial) prediction for each observation, ignoring fixed effects.
    X_index : pd.Index or None
        The index of rows used in X (if needed).
    """
    if newdata is None:
        y_hat = model._Y_untransformed.to_numpy().flatten() - model.resid()
        xfml = model._fml.split("|")[0].split("~")[1]
        X = Formula(xfml).get_model_matrix(model._data)
        X_index = X.index

        coef_idx = np.isin(model._coefnames, X.columns)
        X = X[np.array(model._coefnames)[coef_idx]]
        X = X.to_numpy()
        return y_hat, X, X_index

    else:
        # Convert newdata to a Pandas DataFrame if needed
        newdata = _narwhals_to_pandas(newdata).reset_index(drop=False)

        if not model._X_is_empty:
            if model._icovars is not None:
                raise NotImplementedError(
                    "predict() with argument newdata is not supported with i() syntax."
                )

            if hasattr(model, "_model_spec") and model._model_spec is not None:
                rhs_spec = model._model_spec.fml_second_stage.rhs
                X = rhs_spec.get_model_matrix(newdata, context=context)
            else:
                xfml = model._fml.split("|")[0].split("~")[1]
                X = Formula(xfml).get_model_matrix(newdata, context=context)

            X_index = X.index

            coef_idx = np.isin(model._coefnames, X.columns)
            X = X[np.array(model._coefnames)[coef_idx]]
            X = X.to_numpy()

            # Initialize y_hat with NaNs, fill in only for valid (X_index) rows
            y_hat = np.full(newdata.shape[0], np.nan)
            y_hat[X_index] = X @ model._beta_hat[coef_idx]
        else:
            # If no X in the model, predictions start at zero
            X = None
            X_index = None
            y_hat = np.zeros(newdata.shape[0])

        return y_hat, X, X_index


def _get_fixed_effects_prediction_component(
    model, newdata: pd.DataFrame, atol: float = 1e-6, btol: float = 1e-6
):
    """
    Compute the fixed effect contribution to the prediction.

    Parameters
    ----------
    model : Feols
        The fitted Feols model.
    newdata : pd.DataFrame
        Data for predictions.
    atol : float
    btol : float

    Returns
    -------
    np.ndarray
        fe_hat: the sum of fixed effects contributions for each observation.
    """
    fe_hat = np.zeros(newdata.shape[0])

    if model._has_fixef:
        if model._sumFE is None:
            model.fixef(atol, btol)

        fvals = model._fixef.split("+")

        # warn if newdata types do not match
        mismatched_fixef_types = [
            x for x in fvals if newdata[x].dtypes != model._data[x].dtypes
        ]
        if mismatched_fixef_types:
            warnings.warn(
                f"Data types of fixed effects {mismatched_fixef_types} do not match "
                "the model data. This leads to possible mismatched keys in the fixed "
                "effect dictionary -> NaN predictions for new levels."
            )

        df_fe = newdata[fvals].astype(str)

        # populate fixed effect dicts with omitted categories handling
        fixef_dicts = {}
        for f in fvals:
            fdict = model._fixef_dict[f"C({f})"]
            omitted_cat = set(model._data[f].unique().astype(str).tolist()) - set(
                fdict.keys()
            )
            if omitted_cat:
                fdict.update({x: 0 for x in omitted_cat})
            fixef_dicts[f"C({f})"] = fdict

        _fixef_mat = _apply_fixef_numpy(df_fe.values, fixef_dicts)
        fe_hat += np.sum(_fixef_mat, axis=1)

    return fe_hat


def _apply_fixef_numpy(df_fe_values, fixef_dicts):
    fixef_mat = np.zeros_like(df_fe_values, dtype=float)
    for i, (_, subdict) in enumerate(fixef_dicts.items()):
        unique_levels, inverse = np.unique(df_fe_values[:, i], return_inverse=True)
        mapping = np.array([subdict.get(level, np.nan) for level in unique_levels])
        fixef_mat[:, i] = mapping[inverse]

    return fixef_mat


def _get_prediction_se(model, X: np.ndarray) -> np.ndarray:
    """
    Compute prediction standard error for each row in X.

    Parameters
    ----------
    model : Feols
        The fitted Feols model.
    X : np.ndarray
        The design matrix for newdata.

    Returns
    -------
    se : np.ndarray
        The prediction standard error for each observation.
    """
    return np.sqrt(np.einsum("ij,jk,ik->i", X, model._vcov, X))


def _compute_prediction_error(
    model, nobs: int, yhat: np.ndarray, X: np.ndarray, X_index: np.ndarray, alpha: float
) -> pd.DataFrame:
    """
    Fill a DataFrame with predictions and confidence intervals.

    Parameters
    ----------
    model : Feols
        The fitted Feols model.
    nobs : int
        The number of rows in the prediction DataFrame.
    yhat : np.ndarray
        The predicted values.
    X : np.ndarray
        The design matrix.
    X_index : np.ndarray
        The index of rows used in X.
    alpha : float
        The confidence level.

    Returns
    -------
    prediction_df : pd.DataFrame
        The DataFrame with predictions, prediction SEs and confidence intervals.
    """
    columns = ["fit", "se_fit", "ci_low", "ci_high"]

    prediction_df = pd.DataFrame(np.nan, index=range(nobs), columns=columns)

    prediction_df["fit"] = yhat
    prediction_df.loc[X_index, "se_fit"] = _get_prediction_se(model=model, X=X)
    z_crit = t.ppf(1 - alpha / 2, model._N - model._k)
    sigma2 = np.sum(model._u_hat**2) / (model._N - model._k)
    prediction_df.loc[X_index, "ci_low"] = prediction_df["fit"] - z_crit * np.sqrt(
        prediction_df["se_fit"] ** 2 + sigma2
    )
    prediction_df.loc[X_index, "ci_high"] = prediction_df["fit"] + z_crit * np.sqrt(
        prediction_df["se_fit"] ** 2 + sigma2
    )

    return prediction_df



import numpy as np
import pandas as pd


def get_blw():
    """DGP for effect heterogeneity in panel data from Baker, Larcker, and Wang (2022)."""
    n = np.arange(1, 31)
    id_ = np.arange(1, 1001)
    blw = pd.DataFrame(
        [(n_, id__) for n_ in n for id__ in id_], columns=["n", "id"]
    ).eval(
        """
            year = n + 1980 - 1
            state = 1 + (id - 1) // 25
            group = 1 + (state - 1) // 10
            treat_date = 1980 + group * 6
            time_til = year - treat_date
            treat = time_til >= 0
        """
    )
    blw["firms"] = np.random.uniform(0, 5, size=len(blw))
    blw["e"] = np.random.normal(0, 0.5**2, size=len(blw))
    blw["te"] = np.random.normal(10 - 2 * (blw["group"] - 1), 0.2**2, size=len(blw))
    blw.eval(
        """
        y = firms + n + treat * te * (year - treat_date + 1) + e
        y2 = firms + n + te * treat + e
    """,
        inplace=True,
    )
    return blw


# convert this into a single assignment

treat_effect_vector_1 = np.log(2 * np.arange(1, 30 - 15 + 1))
treat_effect_vector_1[8:] = 0


def get_sharkfin(
    num_units=1000,
    num_periods=30,
    num_treated=200,
    treatment_start=15,
    hetfx=False,
    base_treatment_effect=None,
    sigma_unit=2,
    sigma_time=1,
    sigma_epsilon=0.5,
    ar_coef=0.8,
    het_AR=False,
    return_dataframe=True,
    seed=42,
):
    """Sharkfin panel data DGP with heterogeneous treatment effects and effect reversal.

    Args:
        num_units (int, optional): Defaults to 100.
        num_periods (int, optional): Defaults to 30.
        num_treated (int, optional): Defaults to 50.
        treatment_start (int, optional): Defaults to 15.
        hetfx (bool, optional): Heterogeneous effects. Defaults to True.
        base_treatment_effect (_type_, optional): _description_. Defaults to 0.1*np.log(np.arange(1, 30 - 15 + 1)).
        return_dataframe (bool, optional): _description_. Defaults to True.
        sigma_unit (int, optional): _description_. Defaults to 1.
        sigma_time (float, optional): _description_. Defaults to 0.5.
        sigma_epsilon (float, optional): _description_. Defaults to 0.5.
        het_AR (bool, optional): _description_. Defaults to False.
    """
    np.random.seed(seed)
    if base_treatment_effect is None:
        base_treatment_effect = np.where(
            np.arange(1, 30 - 15 + 1) <= 8,
            0.2 * np.log(2 * np.arange(1, 30 - 15 + 1)),
            0,
        )
    unit_intercepts = np.random.normal(0, sigma_unit, num_units)

    # Generate day-of-the-week pattern
    day_effects = np.array(
        [-0.1, 0.1, 0, 0, 0.1, 0.5, 0.5]
    )  # Stronger effects on weekends
    day_pattern = np.tile(day_effects, num_periods // 7 + 1)[:num_periods]

    # Generate autoregressive structure
    ar_coef_time = 0.2
    ar_noise_time = np.random.normal(0, sigma_time, num_periods)
    time_intercepts = np.zeros(num_periods)
    time_intercepts[0] = ar_noise_time[0]
    for t in range(1, num_periods):
        time_intercepts[t] = ar_coef_time * time_intercepts[t - 1] + ar_noise_time[t]
    # Combine day-of-the-week pattern and autoregressive structure
    time_intercepts = day_pattern + time_intercepts - np.mean(time_intercepts)
    # Generate autoregressive noise for each unit
    ar_noise = np.random.normal(0, sigma_epsilon, (num_units, num_periods))
    if het_AR:
        ar_coef = np.random.normal(ar_coef, 0.1, num_units)
    noise = np.zeros((num_units, num_periods))
    noise[:, 0] = ar_noise[:, 0]
    for t in range(1, num_periods):
        noise[:, t] = ar_coef * noise[:, t - 1] + ar_noise[:, t]
    # N X T matrix of potential outcomes under control
    Y0 = unit_intercepts[:, np.newaxis] + time_intercepts[np.newaxis, :] + noise
    # Generate the base treatment effect (concave structure)
    # Generate heterogeneous multipliers for each unit
    if hetfx:
        heterogeneous_multipliers = np.random.uniform(0.5, 1.5, num_units)
    else:
        heterogeneous_multipliers = np.ones(num_units)

    # Create a 2D array to store the heterogeneous treatment effects
    treatment_effect = np.zeros((num_units, num_periods - treatment_start))
    for i in range(num_units):
        treatment_effect[i, :] = heterogeneous_multipliers[i] * base_treatment_effect

    # random assignment
    treated_units = np.random.choice(num_units, num_treated, replace=False)
    treatment_status = np.zeros((num_units, num_periods), dtype=bool)
    treatment_status[treated_units, treatment_start:] = True

    # Apply the heterogeneous treatment effect to the treated units
    Y1 = Y0.copy()
    for t in range(treatment_start, num_periods):
        Y1[:, t][treatment_status[:, t]] += treatment_effect[:, t - treatment_start][
            treatment_status[:, t]
        ]

    result = {
        "Y1": Y1,
        "Y0": Y0,
        "W": treatment_status,
        "unit_intercepts": unit_intercepts,
        "time_intercepts": time_intercepts,
    }

    if return_dataframe:
        # Create a DataFrame
        unit_ids = np.repeat(np.arange(num_units), num_periods)
        time_ids = np.tile(np.arange(num_periods), num_units)
        W_it = treatment_status.flatten()
        Y_it = np.where(W_it, Y1.flatten(), Y0.flatten())
        df = pd.DataFrame(
            {
                "unit": unit_ids,
                "year": time_ids,
                "treat": W_it.astype(int),
                "Y": Y_it,
            }
        )
        # assign units to ever treated if the max of W_it is 1
        df["ever_treated"] = df.groupby("unit")["treat"].transform("max")
        result["dataframe"] = df
        return df
    return result


def get_panel_dgp_stagg(
    num_units=1_000,
    num_periods=30,
    num_treated=None,
    treatment_start_cohorts=None,
    sigma_unit=1,
    sigma_time=0.5,
    sigma_epsilon=0.2,
    hetfx=False,
    base_treatment_effects=None,
    return_dataframe=True,
    ar_coef=0.8,
):
    """Panel DGP with staggered treatment effects and effect heterogeneity."""
    if num_treated is None:
        num_treated = [250, 500, 150]
    if treatment_start_cohorts is None:
        treatment_start_cohorts = [10, 15, 20]
    if base_treatment_effects is None:
        # Cohort 1: mean reversal: big bump that decays to zero within 10 days, then zero
        # Cohort 2: shark-fin - logarithmic for the first week, then 0
        # Cohort 3: sinusoidal
        # effect functions
        treat_effect_vector_1 = np.log(
            2 * np.arange(1, num_periods - treatment_start_cohorts[1] + 1)
        )
        treat_effect_vector_1[8:] = 0  # switch off effects after a week
        base_treatment_effects = [
            np.r_[
                np.linspace(2, 0, num_periods - treatment_start_cohorts[0] - 10),
                np.repeat(0, 10),
            ],
            treat_effect_vector_1,
            np.sin(
                np.arange(1, num_periods - treatment_start_cohorts[2] + 1)
            ),  # Treatment effect function for cohort 2
        ]
    # unit FEs
    unit_intercepts = np.random.normal(0, sigma_unit, num_units)
    ####################################################################
    # time FEs: Generate day-of-the-week pattern
    day_effects = np.array(
        [-0.1, 0.1, 0, 0, 0.1, 0.5, 0.5]
    )  # Stronger effects on weekends
    day_pattern = np.tile(day_effects, num_periods // 7 + 1)[:num_periods]
    # autoregressive structure in time FEs
    ar_coef_time = 0.2
    ar_noise_time = np.random.normal(0, sigma_time, num_periods)
    time_intercepts = np.zeros(num_periods)
    time_intercepts[0] = ar_noise_time[0]
    for t in range(1, num_periods):
        time_intercepts[t] = ar_coef_time * time_intercepts[t - 1] + ar_noise_time[t]
    # Combine day-of-the-week pattern and autoregressive structure
    time_intercepts = day_pattern + time_intercepts - np.mean(time_intercepts)
    ####################################################################
    # Generate autoregressive noise for each unit
    ar_noise = np.random.normal(0, sigma_epsilon, (num_units, num_periods))
    noise = np.zeros((num_units, num_periods))
    noise[:, 0] = ar_noise[:, 0]
    for t in range(1, num_periods):
        noise[:, t] = ar_coef * noise[:, t - 1] + ar_noise[:, t]
    # N X T matrix of potential outcomes under control
    Y0 = unit_intercepts[:, np.newaxis] + time_intercepts[np.newaxis, :] + noise
    ####################################################################
    # Generate heterogeneous multipliers for each unit
    if hetfx:
        heterogeneous_multipliers = np.random.uniform(0.5, 1.5, num_units)
    else:
        heterogeneous_multipliers = np.ones(num_units)
    # random assignment
    treated_units = np.array([], dtype=int)
    treatment_status = np.zeros((num_units, num_periods), dtype=bool)
    ####################################################################
    # Create a 2D array to store the heterogeneous treatment effects
    treatment_effect = np.zeros((num_units, num_periods))
    # iterate over treatment cohorts
    for cohort_idx, (treatment_start, num_treated_cohort) in enumerate(
        zip(treatment_start_cohorts, num_treated)
    ):
        base_treatment_effect = base_treatment_effects[cohort_idx]
        cohort_treatment_effect = np.zeros((num_units, num_periods - treatment_start))

        for i in range(num_units):
            cohort_treatment_effect[i, :] = (
                heterogeneous_multipliers[i] * base_treatment_effect
            )
        cohort_treated_units = np.random.choice(
            np.setdiff1d(np.arange(num_units), treated_units),
            num_treated_cohort,
            replace=False,
        )
        treated_units = np.concatenate((treated_units, cohort_treated_units))
        treatment_status[cohort_treated_units, treatment_start:] = True
        treatment_effect[cohort_treated_units, treatment_start:] += (
            cohort_treatment_effect[cohort_treated_units, :]
        )

    # Apply the heterogeneous treatment effect to the treated units
    Y1 = Y0.copy()
    Y1[treatment_status] += treatment_effect[treatment_status]
    ####################################################################
    result = {
        "Y1": Y1,
        "Y0": Y0,
        "W": treatment_status,
        "unit_intercepts": unit_intercepts,
        "time_intercepts": time_intercepts,
    }

    if return_dataframe:
        # Create a DataFrame
        unit_ids = np.repeat(np.arange(num_units), num_periods)
        time_ids = np.tile(np.arange(num_periods), num_units)
        W_it = treatment_status.flatten().astype(int)
        Y_it = np.where(W_it, Y1.flatten(), Y0.flatten())
        unit_intercepts_flat = np.repeat(unit_intercepts, num_periods)
        time_intercepts_flat = np.tile(time_intercepts, num_units)
        df = pd.DataFrame(
            {
                "unit_id": unit_ids,
                "time_id": time_ids,
                "W_it": W_it,
                "Y_it": Y_it,
                "unit_intercept": unit_intercepts_flat,
                "time_intercept": time_intercepts_flat,
            }
        )
        result["dataframe"] = df
    return result


def gelbach_data(nobs):
    "Create data for testing of Gelbach Decomposition."
    rng = np.random.default_rng(49392)
    df = pd.DataFrame(index=range(nobs))
    df["x1"] = rng.normal(size=nobs)

    df["x21"] = df["x1"] * 1 + rng.normal(loc=0, scale=0.1, size=nobs)
    df["x22"] = df["x1"] * 0.25 + df["x21"] * 0.75 + rng.normal(size=nobs)
    df["x23"] = (
        df["x1"] * 0.4
        + df["x21"] * 0.6
        + df["x22"] * 0.4
        + rng.normal(loc=0, scale=0.1, size=nobs)
    )
    df["y"] = (
        df["x1"] * 1
        + df["x21"] * 2
        + df["x22"] * 0.5
        + df["x23"] * 0.75
        + rng.normal(loc=0, scale=0.1, size=nobs)
    )

    return df



import numpy as np
import pandas as pd
import pytest

from pyfixest.estimation.ccv import _compute_CCV
from pyfixest.estimation.estimation import feols


@pytest.fixture
def data(local=False):
    """Load the census data used in the tests."""
    if local:
        return pd.read_stata("C:/Users/alexa/Downloads/census2000_5pc.dta")
    else:
        return pd.read_stata("http://www.damianclarke.net/stata/census2000_5pc.dta")


# function retrieved from Harvard Dataverse
@pytest.mark.extended
def compute_CCV_AAIW(data, depvar, cluster, seed, nmx, pk):
    """
    Compute the CCV variance using a slight variation of AAIW's code.

    The code is based on a Python implementation of the authours
    published under CC0 1.0 Deed and available at
    https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/27VMOT.
    This function has also benefitted from Daniel Pailanir
    and Damian Clarke's implementation in Stata available at
    https://github.com/Daniel-Pailanir/TSCB-CCV and published under GPL3 License.

    """
    rng = np.random.default_rng(seed)
    data["u"] = rng.choice([False, True], size=data.shape[0])

    u1 = data["u"] == True  # noqa: E712
    u0 = data["u"] == False  # noqa: E712
    w1 = data[nmx] == 1
    w0 = data[nmx] == 0

    # compute alpha, tau using first split
    alpha = data[u1 & w0][depvar].mean()
    tau = data[u1 & w1][depvar].mean() - data[u1 & w0][depvar].mean()

    # compute for each m
    tau_ms = {}
    nm = "tau_"
    pk_term = 0
    for m in data[cluster].unique():
        ind_m = data[cluster] == m
        aux1 = data[u1 & ind_m & w1][depvar].mean()
        aux0 = data[u1 & ind_m & w0][depvar].mean()

        aux1_full = data[ind_m & w1][depvar].mean()
        aux0_full = data[ind_m & w0][depvar].mean()
        aux_tau = aux1 - aux0
        aux_tau_full = aux1_full - aux0_full

        if (np.isnan(aux1)) or (np.isnan(aux0)):
            aux_tau = tau

        aux_nm = nm + str(m)
        tau_ms[aux_nm] = aux_tau

        # compute the pk term in u0
        Nm = data[ind_m].shape[0]
        aux_pk = Nm * ((aux_tau_full - tau) ** 2)
        pk_term = pk_term + aux_pk

    # compute the residuals
    data["resU"] = data[depvar] - alpha - data[nmx] * tau

    # Wbar
    Wbar = data[u1][nmx].mean()
    # Wbar = data[nmx].mean() # to match Guido

    # pk term
    pk_term = pk_term * (1.0 - pk) / data.shape[0]

    # compute avg Z
    Zavg = (np.sum(data["u"] == False)) / data.shape[0]  # noqa: E712

    # compute the normalized CCV using second split
    n = data.shape[0] * (Wbar**2) * ((1.0 - Wbar) ** 2)

    sum_CCV = 0
    for m in data[cluster].unique():
        ind_m = data[cluster] == m
        data_m = data[u0 & ind_m]
        aux_nm = nm + str(m)

        # tau term
        tau_term = (tau_ms[aux_nm] - tau) * Wbar * (1.0 - Wbar)

        # Residual term
        res_term = (data_m[nmx] - Wbar) * data_m["resU"]

        # square of sums
        sq_sum = np.sum(res_term - tau_term) ** 2

        # sum of squares
        sum_sq = np.sum((res_term - tau_term) ** 2)

        # compute CCV
        sum_CCV += (
            (1.0 / (Zavg**2)) * sq_sum
            - ((1.0 - Zavg) / (Zavg**2)) * sum_sq
            + n * pk_term
        )

    # normalize
    V_CCV = sum_CCV / n
    print("V_CCV: ", V_CCV)

    return V_CCV


@pytest.mark.parametrize("pk", [0.05, 0.5, 0.95])
@pytest.mark.extended
def test_ccv_against_AAIW(data, pk):
    N = data.shape[0]
    Y = data["ln_earnings"].values
    W = data["college"].values
    X = np.concatenate([np.ones((N, 1)), W.reshape(-1, 1)], axis=1)
    cluster_vec = data["state"].values
    seed = 2002

    rng = np.random.default_rng(seed)

    fml = "ln_earnings ~ college"
    tau_full = feols(fml, data=data).coef().xs("college")

    vcov_AAIW = compute_CCV_AAIW(
        data, depvar="ln_earnings", cluster="state", seed=seed, nmx="college", pk=pk
    )
    vcov = _compute_CCV(
        fml=fml,
        X=X,
        Y=Y,
        W=W,
        treatment="college",
        cluster_vec=cluster_vec,
        pk=pk,
        rng=rng,
        data=data,
        tau_full=tau_full,
    )

    assert np.abs(vcov - vcov_AAIW) < 1e-6


@pytest.mark.extended
def test_ccv_internally(data):
    """Test the ccv function internally."""
    # it does not matter where CRV inference is specified
    fit1 = feols("ln_earnings ~ college", data=data)
    fit2 = feols("ln_earnings ~ college", data=data, vcov={"CRV1": "state"})

    res1 = fit1.ccv(
        treatment="college", pk=0.05, qk=1, n_splits=2, seed=929, cluster="state"
    )
    res2 = fit2.ccv(treatment="college", pk=0.05, qk=1, n_splits=2, seed=929)

    assert np.all(res1 == res2)


@pytest.mark.extended
def test_against_stata(data):
    """
    Test the ccv function against the stata implementation of the CCV variance.

    The test values are taken from the readme example of the stata
    implementation of the CCV variance
    and can be found here: https://github.com/Daniel-Pailanir/TSCB-CCV.
    """
    # this can take a while

    fit = feols("ln_earnings ~ college", data=data, vcov={"CRV1": "state"})

    res_ccv1 = fit.ccv(treatment="college", pk=0.05, qk=1, n_splits=4, seed=929).xs(
        "CCV"
    )
    res_ccv2 = fit.ccv(treatment="college", pk=0.5, qk=1, n_splits=4, seed=929).xs(
        "CCV"
    )
    res_ccv3 = fit.ccv(treatment="college", pk=1, qk=0.05, n_splits=4, seed=929).xs(
        "CCV"
    )
    res_ccv4 = fit.ccv(treatment="college", pk=1, qk=0.5, n_splits=4, seed=929).xs(
        "CCV"
    )

    assert np.abs(res_ccv1["2.5%"] - 0.458) < 1e-02
    assert np.abs(res_ccv1["97.5%"] - 0.473) < 1e-02

    assert np.abs(res_ccv2["2.5%"] - 0.458) < 1e-02
    assert np.abs(res_ccv2["97.5%"] - 0.473) < 1e-02

    assert np.abs(res_ccv3["2.5%"] - 0.414) < 1e-02
    assert np.abs(res_ccv3["97.5%"] - 0.517) < 1e-02

    assert np.abs(res_ccv4["2.5%"] - 0.428) < 1e-02
    assert np.abs(res_ccv4["97.5%"] - 0.503) < 1e-02



# Import modules
from pyfixest import (
    did,
    errors,
    estimation,
    report,
    utils,
)
from pyfixest.did import (
    SaturatedEventStudy,
    did2s,
    event_study,
    lpdid,
    panelview,
)

# Import frequently used functions and classes
from pyfixest.estimation import (
    bonferroni,
    feglm,
    feols,
    fepois,
    quantreg,
    rwolf,
    wyoung,
)
from pyfixest.report import coefplot, dtable, etable, iplot, make_table, qplot, summary
from pyfixest.utils import (
    get_data,
    get_ssc,
    ssc,
)

__all__ = [
    "SaturatedEventStudy",
    "bonferroni",
    "coefplot",
    "did",
    "did2s",
    "dtable",
    "errors",
    "estimation",
    "etable",
    "event_study",
    "feglm",
    "feols",
    "fepois",
    "get_data",
    "get_ssc",
    "iplot",
    "lpdid",
    "make_table",
    "panelview",
    "qplot",
    "quantreg",
    "report",
    "rwolf",
    "ssc",
    "summary",
    "utils",
    "wyoung",
]

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("pyfixest")
except PackageNotFoundError:
    __version__ = "unknown"



import duckdb
import numpy as np
import pandas as pd
import pytest
from formulaic.errors import FactorEvaluationError

import pyfixest as pf
from pyfixest.utils.utils import get_data


def test_api():
    df1 = get_data()
    df2 = get_data(model="Fepois")

    fit1 = pf.feols("Y ~ X1 + X2 | f1", data=df1)
    fit2 = pf.estimation.fepois(
        "Y ~ X1 + X2 + f2 | f1", data=df2, vcov={"CRV1": "f1+f2"}
    )
    fit_multi = pf.feols("Y + Y2 ~ X1", data=df2)

    pf.summary(fit1)
    pf.report.summary(fit2)
    pf.etable([fit1, fit2])
    pf.coefplot([fit1, fit2])

    pf.summary(fit_multi)
    pf.etable(fit_multi)
    pf.coefplot(fit_multi)


def test_feols_args():
    """
    Check feols function arguments.

    Arguments to check:
    - copy_data
    - store_data
    - fixef_tol
    - solver
    """
    df = pf.get_data()

    fit1 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, copy_data=True)
    fit2 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, copy_data=False)

    assert (fit1.coef() == fit2.coef()).all()

    fit3 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, store_data=False, fixef_tol=1e-02)
    if hasattr(fit3, "_data"):
        raise AttributeError(
            "The 'fit3' object has the attribute '_data', which should not be present."
        )

    assert fit1.coef().xs("X1") != fit3.coef().xs("X1")
    assert np.abs(fit1.coef().xs("X1") - fit3.coef().xs("X1")) < 0.01

    fit4 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, solver="np.linalg.solve")
    fit5 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, solver="np.linalg.lstsq")

    assert (fit4.coef() == fit5.coef()).all()


def test_fepois_args():
    """
    Check feols function arguments.

    Arguments to check:
    - copy_data
    - store_data
    - fixef_tol
    - solver
    """
    df = pf.get_data(model="Fepois")

    fit1 = pf.fepois(fml="Y ~ X1 | f1 + f2", data=df, copy_data=True)
    fit2 = pf.fepois(fml="Y ~ X1 | f1 + f2", data=df, copy_data=False)

    assert (fit1.coef() == fit2.coef()).all()

    fit3 = pf.fepois(fml="Y ~ X1 | f1 + f2", data=df, store_data=False, fixef_tol=1e-02)
    if hasattr(fit3, "_data"):
        raise AttributeError(
            "The 'fit3' object has the attribute '_data', which should not be present."
        )

    assert fit1.coef().xs("X1") != fit3.coef().xs("X1")
    assert np.abs(fit1.coef().xs("X1") - fit3.coef().xs("X1")) < 0.01

    fit4 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, solver="np.linalg.solve")
    fit5 = pf.feols(fml="Y ~ X1 | f1 + f2", data=df, solver="np.linalg.lstsq")

    np.testing.assert_allclose(fit4.coef(), fit5.coef(), rtol=1e-12)


def test_lean():
    data = pf.get_data()
    fit = pf.feols("Y ~ X1 + X2 | f1", data=data, lean=True)

    assert not hasattr(fit, "_data")
    assert not hasattr(fit, "_X")
    assert not hasattr(fit, "_Y")


def test_duckdb_input():
    data_pandas = pf.get_data()
    data_duckdb = duckdb.query("SELECT * FROM data_pandas")
    fit_pandas = pf.feols("Y ~ X1 | f1 + f2", data=data_pandas)
    fit_duckdb = pf.feols("Y ~ X1 | f1 + f2", data=data_duckdb)
    assert type(fit_pandas) is type(fit_duckdb)
    np.testing.assert_allclose(fit_pandas.coef(), fit_duckdb.coef(), rtol=1e-12)
    np.testing.assert_allclose(fit_pandas.se(), fit_duckdb.se(), rtol=1e-12)


def _lspline(series: pd.Series, knots: list[float]) -> np.array:
    """Generate a linear spline design matrix for the input series based on knots."""
    vector = series.values
    columns = []

    for i, knot in enumerate(knots):
        column = np.minimum(vector, knot if i == 0 else knot - knots[i - 1])
        columns.append(column)
        vector = vector - column

    # Add the remainder as the last column
    columns.append(vector)

    # Combine columns into a design matrix
    return np.column_stack(columns)


@pytest.fixture
def spline_data():
    """Fixture to prepare data with spline splits."""
    data = pf.get_data()
    data["Y"] = np.where(data["Y"] > data["Y"].median(), 1, 0)
    spline_split = _lspline(data["X2"], [0, 1])
    data["X2_0"], data["0_X2_1"], data["1_X2"] = spline_split.T
    return data


@pytest.mark.parametrize(
    "method,family",
    [
        ("feols", None),
        ("feglm", "logit"),
        ("feglm", "probit"),
        ("feglm", "gaussian"),
    ],
)
def test_context_capture(spline_data, method, family):
    method_kwargs = {"data": spline_data}
    if family:
        method_kwargs["family"] = family

    explicit_fit = getattr(pf, method)("Y ~ X2_0 + 0_X2_1 + 1_X2", **method_kwargs)
    context_captured_fit = getattr(pf, method)(
        "Y ~ _lspline(X2,[0,1])", context=0, **method_kwargs
    )
    context_captured_fit_map = getattr(pf, method)(
        "Y ~ _lspline(X2,[0,1])", context={"_lspline": _lspline}, **method_kwargs
    )

    for context_fit in [context_captured_fit, context_captured_fit_map]:
        np.testing.assert_allclose(context_fit.coef(), explicit_fit.coef(), rtol=1e-12)
        np.testing.assert_allclose(context_fit.se(), explicit_fit.se(), rtol=1e-12)

    # FactorEvaluationError for `feols` when context is not set
    if method == "feols":
        with pytest.raises(
            FactorEvaluationError, match="Unable to evaluate factor `_lspline"
        ):
            pf.feols("Y ~ _lspline(X2,[0,1]) | f1 + f2", data=spline_data)



from scipy.stats import norm


def get_hall_sheather_bandwidth(q: float, N: int, alpha: float = 0.05) -> float:
    "Compute the Hall-Sheather bandwidth."
    x0 = norm.ppf(q)
    f0 = norm.pdf(x0)

    h = (
        N ** (-1 / 3)
        * norm.ppf(1 - alpha / 2) ** (2 / 3)
        * ((1.5 * f0**2) / (2 * x0**2 + 1)) ** (1 / 3)
    )

    return h



from typing import Any, Callable, Optional

import numba as nb
import numpy as np
import pandas as pd

from pyfixest.estimation.literals import DemeanerBackendOptions


def demean_model(
    Y: pd.DataFrame,
    X: pd.DataFrame,
    fe: Optional[pd.DataFrame],
    weights: Optional[np.ndarray],
    lookup_demeaned_data: dict[str, Any],
    na_index_str: str,
    fixef_tol: float,
    fixef_maxiter: int,
    demean_func: Callable,
    # demeaner_backend: Literal["numba", "jax", "rust"] = "numba",
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Demean a regression model.

    Demeans a single regression model via the alternating projections algorithm
    (see `demean` function). Prior to demeaning, the function checks if some of
    the variables have already been demeaned and uses values from the cache
    `lookup_demeaned_data` if possible. If the model has no fixed effects, the
    function does not demean the data.

    Parameters
    ----------
    Y : pandas.DataFrame
        A DataFrame of the dependent variable.
    X : pandas.DataFrame
        A DataFrame of the covariates.
    fe : pandas.DataFrame or None
        A DataFrame of the fixed effects. None if no fixed effects specified.
    weights : numpy.ndarray or None
        A numpy array of weights. None if no weights.
    lookup_demeaned_data : dict[str, Any]
        A dictionary with keys for each fixed effects combination and potentially
        values of demeaned data frames. The function checks this dictionary to
        see if some of the variables have already been demeaned.
    na_index_str : str
        A string with indices of dropped columns. Used for caching of demeaned
        variables.
    fixef_tol: float
        The tolerance for the demeaning algorithm.
    fixef_maxiter: int
         The maximum number of iterations for the demeaning algorithm.
    demeaner_backend: DemeanerBackendOptions, optional
        The backend to use for demeaning. Can be either "numba", "jax", or "rust".
        Defaults to "numba".


    Returns
    -------
    tuple[pd.DataFrame, pd.DataFrame, Optional[pd.DataFrame]]
        A tuple of the following elements:
        - Yd : pd.DataFrame
            A DataFrame of the demeaned dependent variable.
        - Xd : pd.DataFrame
            A DataFrame of the demeaned covariates.
        - Id : pd.DataFrame or None
            A DataFrame of the demeaned Instruments. None if no IV.
    """
    YX = pd.concat([Y, X], axis=1)

    yx_names = YX.columns
    YX_array = YX.to_numpy()

    if YX_array.dtype != np.dtype("float64"):
        YX_array = YX_array.astype(np.float64)

    if weights is not None and weights.ndim > 1:
        weights = weights.flatten()

    if fe is not None:
        fe_array = fe.to_numpy()
        # check if looked dict has data for na_index
        if lookup_demeaned_data.get(na_index_str) is not None:
            # get data out of lookup table: list of [algo, data]
            value = lookup_demeaned_data.get(na_index_str)
            if value is not None:
                try:
                    _, YX_demeaned_old = value
                except ValueError:
                    print("Error: Expected the value to be iterable with two elements.")
            else:
                pass

            # get not yet demeaned covariates
            var_diff_names = list(set(yx_names) - set(YX_demeaned_old.columns))

            # if some variables still need to be demeaned
            if var_diff_names:
                # var_diff_names = var_diff_names

                yx_names_list = list(yx_names)
                var_diff_index = [yx_names_list.index(item) for item in var_diff_names]
                # var_diff_index = list(yx_names).index(var_diff_names)
                var_diff = YX_array[:, var_diff_index]
                if var_diff.ndim == 1:
                    var_diff = var_diff.reshape(len(var_diff), 1)

                YX_demean_new, success = demean_func(
                    x=var_diff,
                    flist=fe_array.astype(np.uintp),
                    weights=weights,
                    tol=fixef_tol,
                    maxiter=fixef_maxiter,
                )
                if success is False:
                    raise ValueError(
                        f"Demeaning failed after {fixef_maxiter} iterations."
                    )

                YX_demeaned = pd.DataFrame(YX_demean_new)
                YX_demeaned = np.concatenate([YX_demeaned_old, YX_demean_new], axis=1)
                YX_demeaned = pd.DataFrame(YX_demeaned)

                # check if var_diff_names is a list
                if isinstance(var_diff_names, str):
                    var_diff_names = [var_diff_names]

                YX_demeaned.columns = pd.Index(
                    list(YX_demeaned_old.columns) + var_diff_names
                )

            else:
                # all variables already demeaned
                YX_demeaned = YX_demeaned_old[yx_names]

        else:
            YX_demeaned, success = demean_func(
                x=YX_array,
                flist=fe_array.astype(np.uintp),
                weights=weights,
                tol=fixef_tol,
                maxiter=fixef_maxiter,
            )
            if success is False:
                raise ValueError(f"Demeaning failed after {fixef_maxiter} iterations.")

            YX_demeaned = pd.DataFrame(YX_demeaned)
            YX_demeaned.columns = yx_names

        lookup_demeaned_data[na_index_str] = [None, YX_demeaned]

    else:
        # nothing to demean here
        pass

        YX_demeaned = pd.DataFrame(YX_array)
        YX_demeaned.columns = yx_names

    # get demeaned Y, X (if no fixef, equal to Y, X, I)
    Yd = YX_demeaned[Y.columns]
    Xd = YX_demeaned[X.columns]

    return Yd, Xd


@nb.njit
def _sad_converged(a: np.ndarray, b: np.ndarray, tol: float) -> bool:
    for i in range(a.size):
        if np.abs(a[i] - b[i]) >= tol:
            return False
    return True


@nb.njit(locals=dict(id=nb.uint32))
def _subtract_weighted_group_mean(
    x: np.ndarray,
    sample_weights: np.ndarray,
    group_ids: np.ndarray,
    group_weights: np.ndarray,
    _group_weighted_sums: np.ndarray,
) -> None:
    _group_weighted_sums[:] = 0

    for i in range(x.size):
        id = group_ids[i]
        _group_weighted_sums[id] += sample_weights[i] * x[i]

    for i in range(x.size):
        id = group_ids[i]
        x[i] -= _group_weighted_sums[id] / group_weights[id]


@nb.njit
def _calc_group_weights(
    sample_weights: np.ndarray, group_ids: np.ndarray, n_groups: np.ndarray
):
    n_samples, n_factors = group_ids.shape
    dtype = sample_weights.dtype
    group_weights = np.zeros((n_factors, n_groups), dtype=dtype).T

    for j in range(n_factors):
        for i in range(n_samples):
            id = group_ids[i, j]
            group_weights[id, j] += sample_weights[i]

    return group_weights


@nb.njit(parallel=True)
def demean(
    x: np.ndarray,
    flist: np.ndarray,
    weights: np.ndarray,
    tol: float = 1e-08,
    maxiter: int = 100_000,
) -> tuple[np.ndarray, bool]:
    """
    Demean an array.

    Workhorse for demeaning an input array `x` based on the specified fixed
    effects and weights via the alternating projections algorithm.

    Parameters
    ----------
    x : numpy.ndarray
        Input array of shape (n_samples, n_features). Needs to be of type float.
    flist : numpy.ndarray
        Array of shape (n_samples, n_factors) specifying the fixed effects.
        Needs to already be converted to integers.
    weights : numpy.ndarray
        Array of shape (n_samples,) specifying the weights.
    tol : float, optional
        Tolerance criterion for convergence. Defaults to 1e-08.
    maxiter : int, optional
        Maximum number of iterations. Defaults to 100_000.

    Returns
    -------
    tuple[numpy.ndarray, bool]
        A tuple containing the demeaned array of shape (n_samples, n_features)
        and a boolean indicating whether the algorithm converged successfully.

    Examples
    --------
    ```{python}
    import numpy as np
    import pyfixest as pf
    from pyfixest.utils.dgps import get_blw
    from pyfixest.estimation.demean_ import demean
    from formulaic import model_matrix

    fml = "y ~ treat | state + year"

    data = get_blw()
    data.head()

    Y, rhs = model_matrix(fml, data)
    X = rhs[0].drop(columns="Intercept")
    fe = rhs[1].drop(columns="Intercept")
    YX = np.concatenate([Y, X], axis=1)

    # to numpy
    Y = Y.to_numpy()
    X = X.to_numpy()
    YX = np.concatenate([Y, X], axis=1)
    fe = fe.to_numpy().astype(int)  # demean requires fixed effects as ints!

    YX_demeaned, success = demean(YX, fe, weights = np.ones(YX.shape[0]))
    Y_demeaned = YX_demeaned[:, 0]
    X_demeaned = YX_demeaned[:, 1:]

    print(np.linalg.lstsq(X_demeaned, Y_demeaned, rcond=None)[0])
    print(pf.feols(fml, data).coef())
    ```
    """
    n_samples, n_features = x.shape
    n_factors = flist.shape[1]

    if x.flags.f_contiguous:
        res = np.empty((n_features, n_samples), dtype=x.dtype).T
    else:
        res = np.empty((n_samples, n_features), dtype=x.dtype)

    n_threads = nb.get_num_threads()

    n_groups = flist.max() + 1
    group_weights = _calc_group_weights(weights, flist, n_groups)
    _group_weighted_sums = np.empty((n_threads, n_groups), dtype=x.dtype)

    x_curr = np.empty((n_threads, n_samples), dtype=x.dtype)
    x_prev = np.empty((n_threads, n_samples), dtype=x.dtype)

    not_converged = 0
    for k in nb.prange(n_features):
        tid = nb.get_thread_id()

        xk_curr = x_curr[tid, :]
        xk_prev = x_prev[tid, :]
        for i in range(n_samples):
            xk_curr[i] = x[i, k]
            xk_prev[i] = x[i, k] - 1.0

        for _ in range(maxiter):
            for j in range(n_factors):
                _subtract_weighted_group_mean(
                    xk_curr,
                    weights,
                    flist[:, j],
                    group_weights[:, j],
                    _group_weighted_sums[tid, :],
                )
            if _sad_converged(xk_curr, xk_prev, tol):
                break

            xk_prev[:] = xk_curr[:]
        else:
            not_converged += 1

        res[:, k] = xk_curr[:]

    success = not not_converged
    return (res, success)


def _set_demeaner_backend(
    demeaner_backend: DemeanerBackendOptions,
) -> Callable:
    """Set the demeaning backend.

    Currently, we allow for a numba backend, rust backend and a jax backend.
    The latter is expected to be faster on GPU.

    Parameters
    ----------
    demeaner_backend : Literal["numba", "jax", "rust"]
        The demeaning backend to use.

    Returns
    -------
    Callable
        The demeaning function.

    Raises
    ------
    ValueError
        If the demeaning backend is not supported.
    """
    if demeaner_backend == "rust":
        from pyfixest.core.demean import demean as demean_rs

        return demean_rs
    elif demeaner_backend == "numba":
        return demean
    elif demeaner_backend == "jax":
        from pyfixest.estimation.jax.demean_jax_ import demean_jax

        return demean_jax
    else:
        raise ValueError(f"Invalid demeaner backend: {demeaner_backend}")



import warnings
from typing import Union

import numpy as np
import pandas as pd
from scipy.stats import t

from pyfixest.estimation.feiv_ import Feiv
from pyfixest.estimation.feols_ import Feols
from pyfixest.estimation.fepois_ import Fepois
from pyfixest.estimation.FixestMulti_ import FixestMulti
from pyfixest.report.summarize import _post_processing_input_checks

ModelInputType = Union[FixestMulti, list[Union[Feols, Fepois, Feiv]]]


def bonferroni(models: ModelInputType, param: str) -> pd.DataFrame:
    """
    Compute Bonferroni adjusted p-values for multiple hypothesis testing.

    For each model, it is assumed that tests to adjust are of the form
    "param = 0".

    Parameters
    ----------
    models : A supported model object (Feols, Fepois, Feiv, FixestMulti) or a list of
            Feols, Fepois & Feiv models.
    param : str
        The parameter for which the p-values should be adjusted.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing estimation statistics, including the Bonferroni
        adjusted p-values.

    Examples
    --------
    ```{python}
    import pyfixest as pf
    from pyfixest.utils import get_data

    data = get_data().dropna()
    fit1 = pf.feols("Y ~ X1", data=data)
    fit2 = pf.feols("Y ~ X1 + X2", data=data)
    bonf_df = pf.bonferroni([fit1, fit2], param="X1")
    bonf_df
    ```
    """
    models = _post_processing_input_checks(models)
    all_model_stats = pd.DataFrame()
    S = len(models)
    pvalues = np.zeros(S)
    for i, model in enumerate(models):
        if param not in model._coefnames:
            raise ValueError(
                f"Parameter '{param}' not found in the model {model._fml}."
            )
        pvalues[i] = model.pvalue().xs(param)
        all_model_stats = pd.concat([all_model_stats, model.tidy().xs(param)], axis=1)

    adjusted_pvalues = np.minimum(1, pvalues * S)

    all_model_stats.loc["Bonferroni Pr(>|t|)"] = adjusted_pvalues
    all_model_stats.columns = pd.Index([f"est{i}" for i, _ in enumerate(models)])

    return all_model_stats


def rwolf(
    models: ModelInputType,
    param: str,
    reps: int,
    seed: int,
    sampling_method: str = "wild-bootstrap",
) -> pd.DataFrame:
    """
    Compute Romano-Wolf adjusted p-values for multiple hypothesis testing.

    For each model, it is assumed that tests to adjust are of the form
    "param = 0". This function uses the `wildboottest()` method for running the
    bootstrap, hence models of type `Feiv` or `Fepois` are not supported.

    Parameters
    ----------
    models : list[Feols] or FixestMulti
        A list of models for which the p-values should be computed, or a
        FixestMulti object.
        Models of type `Feiv` or `Fepois` are not supported.
    param : str
        The parameter for which the p-values should be computed.
    reps : int
        The number of bootstrap replications.
    seed : int
        The seed for the random number generator.
    sampling_method : str
        Sampling method for computing resampled statistics.
        Users can choose either bootstrap('wild-bootstrap')
        or randomization inference('ri')

    Returns
    -------
    pd.DataFrame
        A DataFrame containing estimation statistics, including the Romano-Wolf
        adjusted p-values.

    Examples
    --------
    ```{python}
    import pyfixest as pf
    from pyfixest.utils import get_data

    data = get_data().dropna()
    fit = pf.feols("Y ~ Y2 + X1 + X2", data=data)
    pf.rwolf(fit, "X1", reps=9999, seed=123)

    fit1 = pf.feols("Y ~ X1", data=data)
    fit2 = pf.feols("Y ~ X1 + X2", data=data)
    rwolf_df = pf.rwolf([fit1, fit2], "X1", reps=9999, seed=123)

    # use randomization inference - dontrun as too slow
    # rwolf_df = pf.rwolf([fit1, fit2], "X1", reps=9999, seed=123, sampling_method = "ri")

    rwolf_df
    ```
    """
    return _multcomp_resample("rwolf", models, param, reps, seed, sampling_method)


def _get_rwolf_pval(t_stats, boot_t_stats):
    """
    Compute Romano-Wolf adjusted p-values based on bootstrapped(or "ri") t-statistics.

    Parameters
    ----------
    t_stats (np.ndarray): A vector of length S - where S is the number of
                        tested hypotheses - containing the original,
                        non-bootstrappe t-statisics.
    boot_t_stats (np.ndarray): A (B x S) matrix containing the
                            bootstrapped(or "ri") t-statistics.

    Returns
    -------
    np.ndarray: A vector of Romano-Wolf corrected p-values.
    """
    t_stats = np.abs(t_stats)
    boot_t_stats = np.abs(boot_t_stats)

    S = boot_t_stats.shape[1]
    B = boot_t_stats.shape[0]

    pinit = corr_padj = pval = np.zeros(S)
    stepdown_index = np.argsort(t_stats)[::-1]
    ro = np.argsort(stepdown_index)

    for s in range(S):
        if s == 0:
            max_stat = np.max(boot_t_stats, axis=1)
            pinit[s] = min(
                1,
                (np.sum(max_stat >= np.abs(t_stats[stepdown_index[s]])) + 1) / (B + 1),
            )
        else:
            boot_t_stat_udp = np.delete(boot_t_stats, stepdown_index[:s], axis=1)
            max_stat = np.max(boot_t_stat_udp, axis=1)
            pinit[s] = min(
                1,
                (np.sum(max_stat >= np.abs(t_stats[stepdown_index[s]])) + 1) / (B + 1),
            )

    for j in range(S):
        if j == 0:
            corr_padj[j] = pinit[j]
        else:
            corr_padj[j] = max(pinit[j], corr_padj[j - 1])

    # Collect the results
    pval = corr_padj[ro]

    return pval


def wyoung(
    models: ModelInputType,
    param: str,
    reps: int,
    seed: int,
    sampling_method: str = "wild-bootstrap",
) -> pd.DataFrame:
    """
    Compute the Westfall-Young adjusted p-values for multiple hypothesis testing.

    For each model, it is assumed that tests to adjust are of the form
    "param = 0". This function uses the `wildboottest()` method for running the
    bootstrap, hence models of type `Feiv` or `Fepois` are not supported.

    Parameters
    ----------
    models : list[Feols] or FixestMulti
        A list of models for which the p-values should be computed, or a
        FixestMulti object.
        Models of type `Feiv` or `Fepois` are not supported.
    param : str
        The parameter for which the p-values should be computed.
    reps : int
        The number of bootstrap replications.
    seed : int
        The seed for the random number generator.
    sampling_method: str
        Sampling method for computing resampled statistics.
        Users can choose either bootstrap('wild-bootstrap')
        or randomization inference('ri')

    Returns
    -------
    pd.DataFrame
        A DataFrame containing estimation statistics, including the Westfall-Young
        adjusted p-values.

    Examples
    --------
    ```{python}
    from pyfixest.estimation import feols
    from pyfixest.utils import get_data
    from pyfixest.multcomp import wyoung

    data = get_data().dropna()
    fit = feols("Y ~ Y2 + X1 + X2", data=data)
    wyoung(fit, "X1", reps=9999, seed=123)

    fit1 = feols("Y ~ X1", data=data)
    fit2 = feols("Y ~ X1 + X2", data=data)
    wyoung_df = wyoung([fit1, fit2], "X1", reps=9999, seed=123)
    wyoung_df
    ```
    """
    return _multcomp_resample("wyoung", models, param, reps, seed, sampling_method)


def _get_wyoung_pval(p_vals, boot_p_vals):
    """
    Compute Westfall-Young adjusted p-values based on bootstrapped t-statistics.

    Parameters
    ----------
    t_stats (np.ndarray): A vector of length S - where S is the number of
                        tested hypotheses - containing the original,
                        non-bootstrappe t-statisics.
    boot_t_stats (np.ndarray): A (B x S) matrix containing the
                            bootstrapped(or "ri") t-statistics.

    Returns
    -------
    np.ndarray: A vector of Westfall-Young corrected p-values.
    """
    S = boot_p_vals.shape[1]
    B = boot_p_vals.shape[0]

    pinit = corr_padj = pval = np.zeros(S)
    stepdown_index = np.argsort(p_vals)
    ro = np.argsort(stepdown_index)

    # Step 3 (p.28 -- Westfall and Young Free step-down resampling method)
    Qs = np.minimum.accumulate(boot_p_vals[:, stepdown_index[::-1]], axis=1)[:, ::-1]

    # Step 4 and 5
    p_against_null = np.less_equal(Qs, p_vals[stepdown_index])

    pinit = p_against_null.sum(axis=0) / B

    # Step 6: Enforce monotonicity of adjusted p-values
    corr_padj = np.maximum.accumulate(pinit)

    pval = corr_padj[ro]

    return pval


def _multcomp_resample(
    type: str,
    models: ModelInputType,
    param: str,
    reps: int,
    seed: int,
    sampling_method: str = "wild-bootstrap",
) -> pd.DataFrame:
    if isinstance(models, FixestMulti):
        models = models.to_list()

    models = _post_processing_input_checks(models)
    if type not in ["rwolf", "wyoung"]:
        raise ValueError("Type should be one of 'rwolf' and 'wyoung'")
    if sampling_method not in ["wild-bootstrap", "ri"]:
        raise ValueError(
            "Adjustment procedure should be one of 'wild-bootstrap' and 'ri'"
        )

    all_model_stats = pd.DataFrame()
    full_enumeration = False

    S = 0
    for model in models:
        if param not in model._coefnames:
            raise ValueError(
                f"Parameter '{param}' not found in the model {model._fml}."
            )

        if model._is_clustered:
            # model._G a list of length 3
            # for oneway clusering: repeated three times
            G = min(model._G)
            if reps > 2**G:
                warnings.warn(
                    f"""
                              2^(the number of clusters) < the number of boot iterations for at least one model,
                              setting full_enumeration to True and reps = {2**G}.
                              """
                )
                full_enumeration = True

        model_tidy = model.tidy().xs(param)
        all_model_stats = pd.concat([all_model_stats, model_tidy], axis=1)
        S += 1

    t_stats = np.zeros(S)
    boot_t_stats = np.zeros((2**G, S)) if full_enumeration else np.zeros((reps, S))

    _df = np.zeros_like(t_stats)
    p_vals = np.zeros_like(t_stats)
    boot_p_vals = np.zeros_like(boot_t_stats)

    for i in range(S):
        model = models[i]

        if sampling_method == "wild-bootstrap":
            wildboot_res_df, bootstrapped_t_stats = model.wildboottest(
                param=param,
                reps=reps,
                return_bootstrapped_t_stats=True,
                seed=seed,  # all S iterations require the same bootstrap samples, hence seed needs to be reset
            )
            t_stats[i] = wildboot_res_df["t value"]
            boot_t_stats[:, i] = bootstrapped_t_stats

        elif sampling_method == "ri":
            rng = np.random.default_rng(seed)
            model.ritest(
                resampvar=param,
                rng=rng,
                reps=reps,
                type="randomization-t",
                store_ritest_statistics=True,
            )

            t_stats[i] = model._ritest_sample_stat
            boot_t_stats[:, i] = model._ritest_statistics

        if type == "wyoung":
            _df[i] = (
                model._N - model._k
                if model._vcov_type in ["iid", "hetero"]
                else min(model._G) - 1
            )
            p_vals[i] = 2 * (1 - t.cdf(np.abs(t_stats[i]), _df[i]))
            boot_p_vals[:, i] = 2 * (1 - t.cdf(np.abs(boot_t_stats[:, i]), _df[i]))
        elif type == "rwolf":
            pass

    if type == "rwolf":
        pval = _get_rwolf_pval(t_stats, boot_t_stats)
        all_model_stats.loc["RW Pr(>|t|)"] = pval
    elif type == "wyoung":
        pval = _get_wyoung_pval(p_vals, boot_p_vals)
        all_model_stats.loc["WY Pr(>|t|)"] = pval
    else:
        raise ValueError("Invalid adjustment procedure specified")

    all_model_stats.columns = pd.Index([f"est{i}" for i, _ in enumerate(models)])
    return all_model_stats
